import jesse.helpers as jh
import jesse.services.logger as logger
from jesse.store import store


def set_up():
    """

    """
    store.reset()


def test_can_log_error_by_firing_event():
    set_up()

    # fire first error event
    logger.error('first error!!!!!')
    first_logged_error = {'time': jh.now(), 'message': 'first error!!!!!'}

    assert store.logs.errors == [first_logged_error]

    # fire second error event
    logger.error('second error!!!!!')
    second_logged_error = {'time': jh.now(), 'message': 'second error!!!!!'}

    assert store.logs.errors == [first_logged_error, second_logged_error]


def test_can_log_info_by_firing_event():
    set_up()

    # fire first info event
    logger.info('first info!!!!!')
    first_logged_info = {'time': jh.now(), 'message': 'first info!!!!!'}

    assert store.logs.info == [first_logged_info]

    # fire second info event
    logger.info('second info!!!!!')
    second_logged_info = {
        'time': jh.now(),
        'message': 'second info!!!!!'
    }

    assert store.logs.info == [first_logged_info, second_logged_info]



from typing import Union

import numpy as np
import talib


def sarext(candles: np.ndarray, startvalue=0, offsetonreverse=0, accelerationinitlong=0, accelerationlong=0,
           accelerationmaxlong=0, accelerationinitshort=0, accelerationshort=0, accelerationmaxshort=0,
           sequential=False) -> Union[float, np.ndarray]:
    """
    SAREXT - Parabolic SAR - Extended

    :param candles: np.ndarray
    :param startvalue: float - default: 0
    :param offsetonreverse: float - default: 0
    :param accelerationinitlong: float - default: 0
    :param accelerationlong: float - default: 0
    :param accelerationmaxlong: float - default: 0
    :param accelerationinitshort: float - default: 0
    :param accelerationshort: float - default: 0
    :param accelerationmaxshort: float - default: 0
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    res = talib.SAREXT(candles[:, 3], candles[:, 4], startvalue=startvalue, offsetonreverse=offsetonreverse,
                       accelerationinitlong=accelerationinitlong, accelerationlong=accelerationlong,
                       accelerationmaxlong=accelerationmaxlong, accelerationinitshort=accelerationinitshort,
                       accelerationshort=accelerationshort, accelerationmaxshort=accelerationmaxshort)

    return res if sequential else res[-1]



from abc import ABC, abstractmethod


class CandleExchange(ABC):
    """

    """
    def __init__(self, name, count, sleep_time):
        self.name = name
        self.count = count
        self.sleep_time = sleep_time
        self.backup_exchange: CandleExchange = None

    @abstractmethod
    def init_backup_exchange(self):
        pass

    @abstractmethod
    def fetch(self, symbol, start_timestamp):
        """

        :param symbol:
        :param start_timestamp:
        """
        pass

    @abstractmethod
    def get_starting_time(self, symbol) -> int:
        """

        :param symbol:
        """
        pass



from .acosc import acosc
from .ad import ad
from .adosc import adosc
from .adx import adx
from .adxr import adxr
from .alligator import alligator
from .ao import ao
from .apo import apo
from .aroon import aroon
from .aroonosc import aroonosc
from .atr import atr
from .avgprice import avgprice
from .beta import beta
from .bollinger_bands import bollinger_bands
from .bollinger_bands_width import bollinger_bands_width
from .bop import bop
from .cc import cc
from .cci import cci
from .chande import chande
from .cmo import cmo
from .correl import correl
from .cvi import cvi
from .dec_osc import dec_osc
from .decycler import decycler
from .dema import dema
from .di import di
from .dm import dm
from .dx import dx
from .donchian import donchian
from .dpo import dpo
from .efi import efi
from .ema import ema
from .emd import emd
from .emv import emv
from .fisher import fisher
from .fosc import fosc
from .frama import frama
from .gatorosc import gatorosc
from .gauss import gauss
from .hma import hma
from .ht_dcperiod import ht_dcperiod
from .ht_dcphase import ht_dcphase
from .ht_phasor import ht_phasor
from .ht_sine import ht_sine
from .ht_trendline import ht_trendline
from .ht_trendmode import ht_trendmode
from .ichimoku_cloud import ichimoku_cloud
from .itrend import itrend
from .kama import kama
from .keltner import keltner
from .kst import kst
from .kvo import kvo
from .linearreg import linearreg
from .linearreg_angle import linearreg_angle
from .linearreg_intercept import linearreg_intercept
from .linearreg_slope import linearreg_slope
from .lrsi import lrsi
from .macd import macd
from .macdext import macdext
from .mama import mama
from .mass import mass
from .marketfi import marketfi
from .medprice import medprice
from .mfi import mfi
from .midpoint import midpoint
from .midprice import midprice
from .minmax import minmax
from .mom import mom
from .msw import msw
from .natr import natr
from .nvi import nvi
from .obv import obv
from .pattern_recognition import pattern_recognition
from .pivot import pivot
from .ppo import ppo
from .pvi import pvi
from .qstick import qstick
from .roc import roc
from .rocp import rocp
from .rocr import rocr
from .rocr100 import rocr100
from .rsi import rsi
from .sar import sar
from .sarext import sarext
from .sma import sma
from .smma import smma
from .srsi import srsi
from .stddev import stddev
from .stochastic import stoch
from .stochf import stochf
from .supersmoother import supersmoother
from .supertrend import supertrend
from .t3 import t3
from .tema import tema
from .trange import trange
from .trima import trima
from .trix import trix
from .tsf import tsf
from .tsi import tsi
from .typprice import typprice
from .ultosc import ultosc
from .var import var
from .vi import vi
from .vidya import vidya
from .vpt import vpt
from .vwma import vwma
from .vwmacd import vwmacd
from .vosc import vosc
from .wad import wad
from .wclprice import wclprice
from .wilders import wilders
from .willr import willr
from .wma import wma
from .zlema import zlema
from .zscore import zscore



import math
import threading
import time

import arrow
import click
import pydash

import jesse.helpers as jh
from jesse.exceptions import CandleNotFoundInExchange
from jesse.models import Candle
from jesse.modes.import_candles_mode.drivers import drivers
from jesse.modes.import_candles_mode.drivers.interface import CandleExchange


def run(exchange: str, symbol: str, start_date_str: str, skip_confirmation=False):
    """

    :param exchange:
    :param symbol:
    :param start_date_str:
    :param skip_confirmation:
    :return:
    """
    try:
        start_timestamp = jh.arrow_to_timestamp(arrow.get(start_date_str, 'YYYY-MM-DD'))
    except:
        raise ValueError('start_date must be a string representing a date before today. ex: 2020-01-17')

    # more start_date validations
    today = arrow.utcnow().floor('day').timestamp * 1000
    if start_timestamp == today:
        raise ValueError("Today's date is not accepted. start_date must be a string a representing date BEFORE today.")
    elif start_timestamp > today:
        raise ValueError("Future's date is not accepted. start_date must be a string a representing date BEFORE today.")

    click.clear()
    symbol = symbol.upper()

    until_date = arrow.utcnow().floor('day')
    start_date = arrow.get(start_timestamp / 1000)
    days_count = jh.date_diff_in_days(start_date, until_date)
    candles_count = days_count * 1440
    exchange = exchange.title()

    try:
        driver: CandleExchange = drivers[exchange]()
    except KeyError:
        raise ValueError('{} is not a supported exchange'.format(exchange))

    loop_length = int(candles_count / driver.count) + 1
    # ask for confirmation
    if not skip_confirmation:
        click.confirm(
            'Importing {} days candles from "{}" for "{}". Duplicates will be skipped. All good?'
                .format(days_count, exchange, symbol), abort=True, default=True)

    with click.progressbar(length=loop_length, label='Importing candles...') as progressbar:
        for _ in range(candles_count):
            temp_start_timestamp = start_date.timestamp * 1000
            temp_end_timestamp = temp_start_timestamp + (driver.count - 1) * 60000

            # to make sure it won't try to import candles from the future! LOL
            if temp_start_timestamp > jh.now():
                break

            # prevent duplicates calls to boost performance
            count = Candle.select().where(
                Candle.timestamp.between(temp_start_timestamp, temp_end_timestamp),
                Candle.symbol == symbol,
                Candle.exchange == exchange
            ).count()
            already_exists = count == driver.count

            if not already_exists:
                # it's today's candles if temp_end_timestamp < now
                if temp_end_timestamp > jh.now():
                    temp_end_timestamp = arrow.utcnow().floor('minute').timestamp * 1000 - 60000

                # fetch from market
                candles = driver.fetch(symbol, temp_start_timestamp)

                if not len(candles):
                    click.clear()
                    first_existing_timestamp = driver.get_starting_time(symbol)

                    # if driver can't provide accurate get_starting_time()
                    if first_existing_timestamp is None:
                        raise CandleNotFoundInExchange(
                            'No candles exists in the market for this day: {} \n'
                            'Try another start_date'.format(
                                jh.timestamp_to_time(temp_start_timestamp)[:10],
                            )
                        )

                    # handle when there's missing candles during the period
                    if temp_start_timestamp > first_existing_timestamp:
                        # see if there are candles for the same date for the backup exchange,
                        # if so, get those, if not, download from that exchange.
                        driver.init_backup_exchange()
                        if driver.backup_exchange is not None:
                            candles = _get_candles_from_backup_exchange(
                                exchange, driver.backup_exchange, symbol, temp_start_timestamp, temp_end_timestamp
                            )

                    else:
                        if not skip_confirmation:
                            print(jh.color('No candle exists in the market for {}\n'.format(
                                jh.timestamp_to_time(temp_start_timestamp)[:10]), 'yellow'))
                            click.confirm(
                                'First present candle is since {}. Would you like to continue?'.format(
                                    jh.timestamp_to_time(first_existing_timestamp)[:10]
                                )
                                , abort=True, default=True)

                        run(exchange, symbol, jh.timestamp_to_time(first_existing_timestamp)[:10], True)
                        return

                # fill absent candles (if there's any)
                candles = _fill_absent_candles(candles, temp_start_timestamp, temp_end_timestamp)

                # store in the database
                if skip_confirmation:
                    _insert_to_database(candles)
                else:
                    threading.Thread(target=_insert_to_database, args=[candles]).start()

            # add as much as driver's count to the temp_start_time
            start_date = start_date.shift(minutes=driver.count)

            progressbar.update(1)

            # sleep so that the exchange won't get angry at us
            if not already_exists:
                time.sleep(driver.sleep_time)


def _get_candles_from_backup_exchange(
        exchange: str,
        backup_driver: CandleExchange,
        symbol: str,
        start_timestamp: int,
        end_timestamp: int
):
    total_candles = []

    # try fetching from database first
    backup_candles = Candle.select(
        Candle.timestamp, Candle.open, Candle.close, Candle.high, Candle.low,
        Candle.volume
    ).where(
        Candle.timestamp.between(start_timestamp, end_timestamp),
        Candle.exchange == backup_driver.name,
        Candle.symbol == symbol).order_by(Candle.timestamp.asc()).tuples()
    already_exists = len(backup_candles) == (end_timestamp - start_timestamp) / 60_000 + 1
    if already_exists:
        # loop through them and set new ID and exchange
        for c in backup_candles:
            total_candles.append({
                'id': jh.generate_unique_id(),
                'symbol': symbol,
                'exchange': exchange,
                'timestamp': c[0],
                'open': c[1],
                'close': c[2],
                'high': c[3],
                'low': c[4],
                'volume': c[5]
            })

        return total_candles

    # try fetching from market now
    days_count = jh.date_diff_in_days(jh.timestamp_to_arrow(start_timestamp), jh.timestamp_to_arrow(end_timestamp))
    # make sure it's rounded up so that we import maybe more candles, but not less
    if days_count < 1:
        days_count = 1
    if type(days_count) is float and not days_count.is_integer():
        days_count = math.ceil(days_count)
    candles_count = days_count * 1440
    start_date = jh.timestamp_to_arrow(start_timestamp).floor('day')
    for _ in range(candles_count):
        temp_start_timestamp = start_date.timestamp * 1000
        temp_end_timestamp = temp_start_timestamp + (backup_driver.count - 1) * 60000

        # to make sure it won't try to import candles from the future! LOL
        if temp_start_timestamp > jh.now():
            break

        # prevent duplicates
        count = Candle.select().where(
            Candle.timestamp.between(temp_start_timestamp, temp_end_timestamp),
            Candle.symbol == symbol,
            Candle.exchange == backup_driver.name
        ).count()
        already_exists = count == backup_driver.count

        if not already_exists:
            # it's today's candles if temp_end_timestamp < now
            if temp_end_timestamp > jh.now():
                temp_end_timestamp = arrow.utcnow().floor('minute').timestamp * 1000 - 60000

            # fetch from market
            candles = backup_driver.fetch(symbol, temp_start_timestamp)

            if not len(candles):
                raise CandleNotFoundInExchange(
                    'No candles exists in the market for this day: {} \n'
                    'Try another start_date'.format(
                        jh.timestamp_to_time(temp_start_timestamp)[:10],
                    )
                )

            # fill absent candles (if there's any)
            candles = _fill_absent_candles(candles, temp_start_timestamp, temp_end_timestamp)

            # store in the database
            _insert_to_database(candles)

        # add as much as driver's count to the temp_start_time
        start_date = start_date.shift(minutes=backup_driver.count)

        # sleep so that the exchange won't get angry at us
        if not already_exists:
            time.sleep(backup_driver.sleep_time)

    # now try fetching from database again. Why? because we might have fetched more
    # than what's needed, but we only want as much was requested. Don't worry, the next
    # request will probably fetch from database and there won't be any waste!
    backup_candles = Candle.select(
        Candle.timestamp, Candle.open, Candle.close, Candle.high, Candle.low,
        Candle.volume
    ).where(
        Candle.timestamp.between(start_timestamp, end_timestamp),
        Candle.exchange == backup_driver.name,
        Candle.symbol == symbol).order_by(Candle.timestamp.asc()).tuples()
    already_exists = len(backup_candles) == (end_timestamp - start_timestamp) / 60_000 + 1
    if already_exists:
        # loop through them and set new ID and exchange
        for c in backup_candles:
            total_candles.append({
                'id': jh.generate_unique_id(),
                'symbol': symbol,
                'exchange': exchange,
                'timestamp': c[0],
                'open': c[1],
                'close': c[2],
                'high': c[3],
                'low': c[4],
                'volume': c[5]
            })

        return total_candles


def _fill_absent_candles(temp_candles, start_timestamp, end_timestamp):
    if len(temp_candles) == 0:
        raise CandleNotFoundInExchange(
            'No candles exists in the market for this day: {} \n'
            'Try another start_date'.format(
                jh.timestamp_to_time(start_timestamp)[:10],
            )
        )

    symbol = temp_candles[0]['symbol']
    exchange = temp_candles[0]['exchange']
    candles = []
    first_candle = temp_candles[0]
    started = False
    loop_length = ((end_timestamp - start_timestamp) / 60000) + 1

    i = 0
    while i < loop_length:
        candle_for_timestamp = pydash.find(
            temp_candles, lambda c: c['timestamp'] == start_timestamp)

        if candle_for_timestamp is None:
            if started:
                last_close = candles[-1]['close']
                candles.append({
                    'id': jh.generate_unique_id(),
                    'symbol': symbol,
                    'exchange': exchange,
                    'timestamp': start_timestamp,
                    'open': last_close,
                    'high': last_close,
                    'low': last_close,
                    'close': last_close,
                    'volume': 0
                })
            else:
                candles.append({
                    'id': jh.generate_unique_id(),
                    'symbol': symbol,
                    'exchange': exchange,
                    'timestamp': start_timestamp,
                    'open': first_candle['open'],
                    'high': first_candle['open'],
                    'low': first_candle['open'],
                    'close': first_candle['open'],
                    'volume': 0
                })
        # candle is present
        else:
            started = True
            candles.append(candle_for_timestamp)

        start_timestamp += 60000
        i += 1

    return candles


def _insert_to_database(candles):
    Candle.insert_many(candles).on_conflict_ignore().execute()



import numpy as np

import jesse.helpers as jh
from jesse.config import config


class CompletedTrade:
    """A trade is made when a position is opened AND closed."""

    def __init__(self, attributes=None):
        self.id = ''
        self.strategy_name = ''
        self.symbol = ''
        self.exchange = ''
        self.type = ''
        self.timeframe = None
        self.entry_price = np.nan
        self.exit_price = np.nan
        self.take_profit_at = np.nan
        self.stop_loss_at = np.nan
        self.qty = np.nan
        self.orders = []
        self.opened_at = None
        self.closed_at = None
        self.entry_candle_timestamp = None
        self.exit_candle_timestamp = None
        self.reduced_at = None
        self.reduction_timestamp = None
        self.reduction_candle_timestamp = None

        if attributes is None:
            attributes = {}

        for a in attributes:
            setattr(self, a, attributes[a])

    def toJSON(self):
        orders = []
        for o in self.orders:
            orders.append(o.__dict__)
        return {
            "id": self.id,
            "strategy_name": self.strategy_name,
            "symbol": self.symbol,
            "exchange": self.exchange,
            "type": self.type,
            "entry_price": self.entry_price,
            "exit_price": self.exit_price,
            "take_profit_at": self.take_profit_at,
            "stop_loss_at": self.stop_loss_at,
            "qty": self.qty,
            "fee": self.fee,
            "reward": self.reward,
            "size": self.size,
            "risk": self.risk,
            "risk_percentage": self.risk_percentage,
            "R": self.R,
            "PNL": self.PNL,
            "PNL_percentage": self.PNL_percentage,
            "holding_period": self.holding_period,
            "opened_at": self.opened_at,
            "closed_at": self.closed_at,
            "entry_candle_timestamp": self.entry_candle_timestamp,
            "exit_candle_timestamp": self.exit_candle_timestamp,
            "orders": orders,
        }

    def to_dict(self):
        return {
            'id': self.id,
            'strategy_name': self.strategy_name,
            'symbol': self.symbol,
            'exchange': self.exchange,
            'type': self.type,
            'entry_price': self.entry_price,
            'exit_price': self.exit_price,
            'take_profit_at': self.take_profit_at,
            'stop_loss_at': self.stop_loss_at,
            'qty': self.qty,
            'opened_at': self.opened_at,
            'closed_at': self.closed_at,
            'entry_candle_timestamp': self.entry_candle_timestamp,
            'exit_candle_timestamp': self.exit_candle_timestamp,
            "fee": self.fee,
            "reward": self.reward,
            "size": self.size,
            "risk": self.risk,
            "risk_percentage": self.risk_percentage,
            "R": self.R,
            "PNL": self.PNL,
            "PNL_percentage": self.PNL_percentage,
            "holding_period": self.holding_period,
        }

    @property
    def fee(self):
        trading_fee = config['env']['exchanges'][self.exchange]['fee']
        return trading_fee * self.qty * (self.entry_price + self.exit_price)

    @property
    def reward(self):
        return abs(self.take_profit_at - self.entry_price) * self.qty

    @property
    def size(self):
        return self.qty * self.entry_price

    @property
    def risk(self):
        return abs(self.stop_loss_at - self.entry_price) * self.qty

    @property
    def risk_percentage(self):
        return round((self.risk / self.size) * 100, 2)

    @property
    def risk_reward_ratio(self):
        return self.reward / self.risk

    @property
    def R(self):
        """alias for risk_reward_ratio"""
        return self.risk_reward_ratio

    @property
    def PNL(self):
        """PNL"""
        fee = config['env']['exchanges'][self.exchange]['fee']
        return jh.estimate_PNL(self.qty, self.entry_price, self.exit_price,
                               self.type, fee)

    @property
    def PNL_percentage(self):
        """The PNL%"""
        return round((self.PNL / self.size) * 100, 2)

    @property
    def holding_period(self):
        """How many SECONDS has it taken for the trade to be done."""
        return (self.closed_at - self.opened_at) / 1000



import os
import shutil

import jesse.helpers as jh


def generate(name):
    """

    :param name:
    :return:
    """
    path = 'strategies/{}'.format(name)

    # validation for name duplication
    exists = os.path.isdir(path)
    if exists:
        print(jh.color('Strategy "{}" already exists.'.format(name), 'red'))
        return

    # generate from ExampleStrategy
    dirname, filename = os.path.split(os.path.abspath(__file__))

    shutil.copytree('{}/ExampleStrategy'.format(dirname), path)

    # replace 'ExampleStrategy' with the name of the new strategy
    fin = open("{}/__init__.py".format(path), "rt")
    data = fin.read()
    data = data.replace('ExampleStrategy', name)
    fin.close()
    fin = open("{}/__init__.py".format(path), "wt")
    fin.write(data)
    fin.close()

    # output the location of generated strategy directory
    print(jh.color('Strategy created at: {}'.format(path), 'green'))



from collections import namedtuple

import numpy as np
import tulipy as ti

from jesse.helpers import get_candle_source

MSW = namedtuple('MSW', ['sine', 'lead'])

def msw(candles: np.ndarray, period=5, source_type="close", sequential=False) -> MSW:
    """
    MSW - Mesa Sine Wave

    :param candles: np.ndarray
    :param period: int - default: 5
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: MSW(sine, lead)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    msw_sine, msw_lead = ti.msw(np.ascontiguousarray(source), period=period)

    s = np.concatenate((np.full((candles.shape[0] - msw_sine.shape[0]), np.nan), msw_sine), axis=0)
    l = np.concatenate((np.full((candles.shape[0] - msw_lead.shape[0]), np.nan), msw_lead), axis=0)

    if sequential:
        return MSW(s, l)
    else:
        return MSW(s[-1], l[-1])


from typing import Union

import numpy as np
import talib



def dx(candles: np.ndarray, period=14, sequential=False) -> Union[float, np.ndarray]:
    """
    DX - Directional Movement Index

    :param candles: np.ndarray
    :param period: int - default: 14
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    res = talib.DX(candles[:, 3], candles[:, 4], candles[:, 2], timeperiod=period)

    return res if sequential else res[-1]



import jesse.helpers as jh
from jesse.config import config, reset_config
from jesse.enums import exchanges
from jesse.factories import fake_range_candle_from_range_prices
from jesse.modes import backtest_mode
from jesse.routes import router
from jesse.store import store


def get_btc_and_eth_candles():
    """

    :return:
    """
    candles = {}
    candles[jh.key(exchanges.SANDBOX, 'BTCUSD')] = {
        'exchange': exchanges.SANDBOX,
        'symbol': 'BTCUSD',
        'candles': fake_range_candle_from_range_prices(range(101, 200))
    }
    candles[jh.key(exchanges.SANDBOX, 'ETHUSD')] = {
        'exchange': exchanges.SANDBOX,
        'symbol': 'ETHUSD',
        'candles': fake_range_candle_from_range_prices(range(1, 100))
    }
    return candles


def get_btc_candles():
    """

    :return:
    """
    candles = {}
    candles[jh.key(exchanges.SANDBOX, 'BTCUSD')] = {
        'exchange': exchanges.SANDBOX,
        'symbol': 'BTCUSD',
        'candles': fake_range_candle_from_range_prices(range(1, 100))
    }
    return candles


def set_up(routes, fee=0):
    """

    :param routes:
    :param fee:
    """
    reset_config()
    config['env']['exchanges'][exchanges.SANDBOX]['starting_balance'] = 10000
    config['env']['exchanges'][exchanges.SANDBOX]['fee'] = fee
    router.set_routes(routes)
    router.set_extra_candles([])
    store.reset(True)


def test_open_pl_and_total_open_trades():
    set_up([(exchanges.SANDBOX, 'BTCUSD', '1m', 'Test40')])

    backtest_mode.run('2019-04-01', '2019-04-02', get_btc_candles())

    assert len(store.completed_trades.trades) == 0
    assert store.app.total_open_trades == 1
    assert store.app.total_open_pl == 97  # 99 - 2

# def test_statistics_for_trades_without_fee():
#     set_up([
#         (exchanges.SANDBOX, 'ETHUSD', timeframes.MINUTE_5, 'Test06'),
#     ])
#
#     candles = {}
#     key = jh.key(exchanges.SANDBOX, 'ETHUSD')
#     candles[key] = {
#         'exchange': exchanges.SANDBOX,
#         'symbol': 'ETHUSD',
#         'candles': test_candles_1
#     }
#
#     # run backtest (dates are fake just to pass)
#     backtest_mode.run('2019-04-01', '2019-04-02', candles)
#     assert len(store.completed_trades.trades) == 2
#     stats_trades = stats.trades(store.completed_trades.trades, store.app.daily_balance)
#
#     assert stats_trades == {
#         'total': 2,
#         'starting_balance': 10000,
#         'finishing_balance': 10004.7,
#         'win_rate': 0.5,
#         'ratio_avg_win_loss': 1.47,
#         'longs_count': 1,
#         'longs_percentage': 50.0,
#         'short_percentage': 50.0,
#         'shorts_count': 1,
#         'fee': 0,
#         'net_profit': 4.7,
#         'net_profit_percentage': 0.05,
#         'sharpe_ratio': np.nan,
#         'average_win': 14.7,
#         'average_loss': 10,
#         'expectancy': 2.35,
#         'expectancy_percentage': 0.02,
#         'expected_net_profit_every_100_trades': 2,
#         'average_holding_period': 960.0,
#         'average_losing_holding_period': 1740.0,
#         'average_winning_holding_period': 180.0,
#         'gross_loss': -10,
#         'gross_profit': 14.7,
#         'max_drawdown': 0,
#         'annual_return': 6.1
#     }
#
#
# def test_stats_for_a_strategy_without_losing_trades():
#     set_up([
#         (exchanges.SANDBOX, 'ETHUSD', timeframes.MINUTE_5, 'Test08'),
#     ])
#
#     candles = {}
#     key = jh.key(exchanges.SANDBOX, 'ETHUSD')
#     candles[key] = {
#         'exchange': exchanges.SANDBOX,
#         'symbol': 'ETHUSD',
#         'candles': test_candles_1
#     }
#
#     # run backtest (dates are fake just to pass)
#     backtest_mode.run('2019-04-01', '2019-04-02', candles)
#     assert len(store.completed_trades.trades) == 1
#     stats_trades = stats.trades(store.completed_trades.trades)
#
#     assert stats_trades == {
#         'total': 1,
#         'starting_balance': 10000,
#         'finishing_balance': 10014.7,
#         'win_rate': 1,
#         'max_R': 1,
#         'min_R': 1,
#         'mean_R': 1,
#         'longs_count': 0,
#         'longs_percentage': 0,
#         'short_percentage': 100,
#         'shorts_count': 1,
#         'fee': 0,
#         'pnl': 14.7,
#         'pnl_percentage': 0.15,
#         'average_win': 14.7,
#         'average_loss': np.nan,
#         'expectancy': 14.7,
#         'expectancy_percentage': 0.15,
#         'expected_pnl_every_100_trades': 15.0,
#         'average_holding_period': 180.0,
#         'average_losing_holding_period': np.nan,
#         'average_winning_holding_period': 180.0
#     }
#
#
# def test_stats_for_a_strategy_without_any_trades():
#     set_up([
#         (exchanges.SANDBOX, 'ETHUSD', timeframes.MINUTE_5, 'Test09'),
#     ])
#
#     candles = {}
#     key = jh.key(exchanges.SANDBOX, 'ETHUSD')
#     candles[key] = {
#         'exchange': exchanges.SANDBOX,
#         'symbol': 'ETHUSD',
#         'candles': test_candles_1
#     }
#
#     # run backtest (dates are fake just to pass)
#     backtest_mode.run('2019-04-01', '2019-04-02', candles)
#     assert len(store.completed_trades.trades) == 0
#     stats_trades = stats.trades(store.completed_trades.trades)
#
#     assert stats_trades == {
#         'total': 0,
#         'starting_balance': 10000,
#         'finishing_balance': 10000,
#         'win_rate': np.nan,
#         'max_R': np.nan,
#         'min_R': np.nan,
#         'mean_R': np.nan,
#         'longs_count': np.nan,
#         'longs_percentage': np.nan,
#         'short_percentage': np.nan,
#         'shorts_count': np.nan,
#         'fee': np.nan,
#         'pnl': np.nan,
#         'pnl_percentage': np.nan,
#         'average_win': np.nan,
#         'average_loss': np.nan,
#         'expectancy': np.nan,
#         'expectancy_percentage': np.nan,
#         'expected_pnl_every_100_trades': np.nan,
#         'average_holding_period': np.nan,
#         'average_winning_holding_period': np.nan,
#         'average_losing_holding_period': np.nan,
#     }
#
# #
# #
# # def test_statistics_for_trades_with_fee():
# #     set_up([
# #         (exchanges.SANDBOX, 'ETHUSD', timeframes.MINUTE_5, 'Test06'),
# #     ], 0.002)
# #
# #     candles = {}
# #     key = jh.key(exchanges.SANDBOX, 'ETHUSD')
# #     candles[key] = {
# #         'exchange': exchanges.SANDBOX,
# #         'symbol': 'ETHUSD',
# #         'candles': test_candles_1
# #     }
# #
# #     # run backtest (dates are fake just to pass)
# #     backtest_mode('2019-04-01', '2019-04-02', candles)
# #     assert len(store.completed_trades.trades) == 2
# #     stats_trades = stats.trades(store.completed_trades.trades)
# #
# #     assert stats_trades == [
# #         ['total', 2],
# #         ['starting balance', 10000],
# #         ['finishing balance', 9994.29],
# #         ['fee', 10.35],
# #         ['PNL', -5.6514],
# #         ['PNL%', '-0.06%'],
# #         ['PNL% every 100 trades', '-2.83%'],
# #         ['expectancy', -2.83],
# #         ['expectancy%', '-0.03%'],
# #         ['average win', 9.61],
# #         ['average loss', 15.26],
# #         ['win rate', '50%'],
# #         ['min R', 1],
# #         ['average R', 1.5],
# #         ['max R', 2],
# #         ['longs/shorts trades', '50%/50%'],
# #     ]



from jesse.config import config
from jesse.models import Exchange


class ExchangesState:
    """

    """
    def __init__(self):
        self.storage = {}

        for name in config['app']['considering_exchanges']:
            starting_balance = config['env']['exchanges'][name]['starting_balance']
            fee = config['env']['exchanges'][name]['fee']
            self.storage[name] = Exchange(name, starting_balance, fee)



from setuptools import setup, find_packages

VERSION = '0.12.1'
DESCRIPTION = "A trading framework for cryptocurrencies"
REQUIRED_PACKAGES = [
    'psycopg2-binary',
    'pytest',
    'Click',
    'arrow',
    'requests',
    'peewee',
    'pydash',
    'numpy',
    'pandas',
    'tabulate',
    'timeloop',
    'websocket-client',
    'TA-Lib',
    'matplotlib',
    'empyrical',
    'blinker',
    'tulipy'
]

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name='jesse',
    version=VERSION,
    author="Saleh Mirnezami",
    author_email="mirnezami.saleh@gmail.com",
    packages=find_packages(),
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://jesse-ai.com",
    project_urls={
        'Documentation': 'https://docs.jesse-ai.com',
        'Say Thanks!': 'http://forum.jesse-ai.com/',
        'Source': 'http://github.com/jesse-ai/jesse',
        'Tracker': 'https://github.com/jesse-ai/jesse/issues',
    },
    install_requires=REQUIRED_PACKAGES,
    entry_points='''
        [console_scripts]
        jesse=jesse.__init__:cli
    ''',
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    include_package_data=True,
)



class CompletedTrades:
    """

    """
    def __init__(self):
        self.trades = []

    def add_trade(self, trade):
        """

        :param trade:
        """
        self.trades.append(trade)

    @property
    def count(self):
        return len(self.trades)



import threading

import jesse.helpers as jh


class API:
    """

    """
    def __init__(self):
        self.drivers = {}

        if not jh.is_live():
            self.initiate_drivers()

    def initiate_drivers(self):
        for e in jh.get_config('app.considering_exchanges'):
            if jh.is_live():
                def initiate_ws(exchange_name: str):
                    """

                    :param exchange_name:
                    """
                    exchange_class = jh.get_config('app.live_drivers.{}'.format(exchange_name))
                    self.drivers[exchange_name] = exchange_class()

                threading.Thread(target=initiate_ws, args=[e]).start()
            else:
                from jesse.exchanges import Sandbox
                self.drivers[e] = Sandbox(e)

    def market_order(self, exchange, symbol, qty, current_price, side, role, flags):
        """

        :param exchange:
        :param symbol:
        :param qty:
        :param current_price:
        :param side:
        :param role:
        :param flags:
        :return:
        """
        return self.drivers[exchange].market_order(symbol, qty, current_price, side, role, flags)

    def limit_order(self, exchange, symbol, qty, price, side, role, flags):
        """

        :param exchange:
        :param symbol:
        :param qty:
        :param price:
        :param side:
        :param role:
        :param flags:
        :return:
        """
        return self.drivers[exchange].limit_order(symbol, qty, price, side, role, flags)

    def stop_order(self, exchange, symbol, qty, price, side, role, flags):
        """

        :param exchange:
        :param symbol:
        :param qty:
        :param price:
        :param side:
        :param role:
        :param flags:
        :return:
        """
        return self.drivers[exchange].stop_order(symbol, qty, price, side, role, flags)

    def cancel_all_orders(self, exchange, symbol):
        """

        :param exchange:
        :param symbol:
        :return:
        """
        return self.drivers[exchange].cancel_all_orders(symbol)

    def cancel_order(self, exchange, symbol, order_id):
        """

        :param exchange:
        :param symbol:
        :param order_id:
        :return:
        """
        return self.drivers[exchange].cancel_order(symbol, order_id)


api = API()



import pytest

import jesse.services.selectors as selectors
from jesse.config import config, reset_config
from jesse.enums import exchanges
from jesse.exceptions import NegativeBalance
from jesse.store import store


def set_up():
    """

    """
    reset_config()
    config['app']['considering_exchanges'] = [exchanges.SANDBOX]
    config['app']['trading_exchanges'] = [exchanges.SANDBOX]
    config['env']['exchanges'][exchanges.SANDBOX]['starting_balance'] = 2000
    store.reset()


def test_decrease_balance():
    set_up()

    e = selectors.get_exchange(exchanges.SANDBOX)
    assert e.balance == 2000
    e.decrease_balance(None, 100)
    assert e.balance == 1900


def test_increase_balance():
    set_up()

    e = selectors.get_exchange(exchanges.SANDBOX)
    assert e.balance == 2000
    e.increase_balance(None, 100)
    assert e.balance == 2100


def test_negative_balance_validation():
    with pytest.raises(NegativeBalance):
        set_up()

        e = selectors.get_exchange(exchanges.SANDBOX)
        assert e.balance == 2000
        e.decrease_balance(None, 3000)



from collections import namedtuple

import numpy as np

from jesse.helpers import get_candle_source

ITREND = namedtuple('ITREND', ['signal', 'it', 'trigger'])


def itrend(candles: np.ndarray, alpha=0.07, source_type="hl2", sequential=False) -> ITREND:
    """
    Instantaneous Trendline

    :param candles: np.ndarray
    :param alpha: float - default: 0.07
    :param source_type: str - default: "hl2"
    :param sequential: bool - default=False

    :return: ITREND(signal, it, trigger)
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)

    coeff = np.array(
        [(alpha - alpha ** 2 / 4), alpha ** 2 / 2, - (alpha - alpha ** 2 * 3 / 4), 2 * (1 - alpha), - (1 - alpha) ** 2])

    it = np.copy(source)
    for i in range(2, 7):
        it[i] = (source[i] + 2 * source[i - 1] + source[i - 2]) / 4
    for i in range(7, source.shape[0]):
        val = np.array([source[i], source[i - 1], source[i - 2], it[i - 1], it[i - 2]])
        it[i] = np.matmul(coeff, val)

    # compute lead 2 trigger & signal
    lag2 = np.roll(it, 20)
    lag2[:20] = it[:20]
    trigger = 2 * it - lag2
    signal = (trigger > it) * 1 - (trigger < it) * 1

    if sequential:
        return ITREND(signal, it, trigger)
    else:
        return ITREND(signal[-1], it[-1], trigger[-1])



import jesse.services.selectors as selectors
from jesse.config import config, reset_config
from jesse.enums import exchanges
from jesse.models import Position
from jesse.store import store


def set_up():
    """

    """
    reset_config()
    config['app']['considering_exchanges'] = [exchanges.SANDBOX]
    config['env']['exchanges'][exchanges.SANDBOX]['fee'] = 0
    config['env']['exchanges'][exchanges.SANDBOX]['starting_balance'] = 1000
    config['app']['trading_exchanges'] = [exchanges.SANDBOX]
    config['app']['trading_symbols'] = ['BTCUSD']
    config['app']['trading_timeframes'] = ['5m']
    store.reset()


def test_close_position():
    set_up()

    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 50,
        'qty': 2,
    })
    e = selectors.get_exchange('Sandbox')
    assert p.exit_price is None

    p._close(50)

    assert p.qty == 0
    assert e.balance == 1100
    assert p.entry_price is None
    assert p.exit_price == 50


def test_increase_a_long_position():
    set_up()

    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 50,
        'qty': 2,
    })
    e = selectors.get_exchange('Sandbox')

    p._increase(2, 100)

    assert p.qty == 4
    assert p.entry_price == 75
    assert e.balance == 800


def test_increase_a_short_position():
    set_up()

    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 50,
        'qty': -2,
    })
    e = selectors.get_exchange('Sandbox')

    p._increase(2, 40)

    assert p.qty == -4
    assert p.entry_price == 45
    assert e.balance == 920


def test_initiating_position():
    position = Position(exchanges.SANDBOX, 'BTCUSD', {
        'current_price': 100,
        'qty': 0
    })

    assert position.exchange_name == 'Sandbox'
    assert position.symbol == 'BTCUSD'
    assert position.current_price == 100
    assert position.qty == 0
    assert position.closed_at is None
    assert position.opened_at is None
    assert position.entry_price is None
    assert position.exit_price is None


def test_is_able_to_close_via_reduce_postion_too():
    set_up()

    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 50,
        'qty': 2,
    })
    e = selectors.get_exchange('Sandbox')

    p._reduce(2, 50)

    assert p.qty == 0
    assert e.balance == 1100


def test_open_position():
    set_up()

    p = Position(exchanges.SANDBOX, 'BTCUSD')
    e = selectors.get_exchange('Sandbox')

    assert p.qty == 0
    assert p.entry_price is None
    assert p.exit_price is None
    assert p.current_price is None
    assert e.balance == 1000

    p._open(1, 50)

    assert p.qty == 1
    assert p.entry_price == 50
    assert p.exit_price is None
    assert e.balance == 950
def test_position_is_close():
    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 60,
        'qty': 0,
    })
    assert p.is_close is True

    p.qty = 2
    assert p.is_close is False


def test_position_is_open():
    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 60,
        'qty': 2,
    })
    assert p.is_open is True

    p.qty = 0
    assert p.is_open is False


def test_position_pnl():
    # long winning position
    p1: Position = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 100,
        'current_price': 110,
        'qty': 2,
    })
    assert p1.pnl == 20

    # long losing position
    p2: Position = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 100,
        'current_price': 90,
        'qty': 2,
    })
    assert p2.pnl == -20

    # short winning position
    p3: Position = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 100,
        'current_price': 90,
        'qty': -2,
    })
    assert p3.pnl == 20

    # short losing position
    p3: Position = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 100,
        'current_price': 110,
        'qty': -2,
    })
    assert p3.pnl == -20


def test_position_pnl_percentage():
    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 60,
        'qty': 2,
    })

    # long position
    assert p.pnl_percentage == 20

    p.current_price -= 20
    assert p.pnl_percentage == -20

    # short position
    p.entry_price = 50
    p.qty = -2
    p.current_price = 40
    assert p.pnl_percentage == 20


def test_position_type():
    p = Position(exchanges.SANDBOX, 'BTCUSD', {'current_price': 100, 'qty': 0})
    assert p.type == 'close'

    p = Position(exchanges.SANDBOX, 'BTCUSD', {'current_price': 100, 'qty': 1})
    assert p.type == 'long'

    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'current_price': 100,
        'qty': -1
    })
    assert p.type == 'short'


def test_position_value():
    long_position = Position(exchanges.SANDBOX, 'BTCUSD', {'current_price': 100, 'qty': 1})
    short_position = Position(exchanges.SANDBOX, 'BTCUSD', {'current_price': 100, 'qty': -1})

    assert long_position.value == 100
    assert short_position.value == 100


def test_reduce_a_long_position():
    set_up()

    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 50,
        'qty': 2,
    })
    e = selectors.get_exchange('Sandbox')

    p._reduce(1, 50)

    assert p.qty == 1
    assert e.balance == 1050


def test_reduce_a_short_position():
    set_up()

    p = Position(exchanges.SANDBOX, 'BTCUSD', {
        'entry_price': 50,
        'current_price': 50,
        'qty': -2,
    })
    e = selectors.get_exchange('Sandbox')

    p._reduce(1, 50)

    assert p.qty == -1
    assert e.balance == 1050





class Route:
    """

    """
    def __init__(self, exchange, symbol, timeframe=None, strategy_name=None, dna=None):
        self.exchange = exchange
        self.symbol = symbol
        self.timeframe = timeframe
        self.strategy_name = strategy_name
        self.strategy = None
        self.dna = dna



config = {
    # these values are related to the user's environment
    'env': {
        'databases': {
            'postgres_host': '127.0.0.1',
            'postgres_name': 'jesse_db',
            'postgres_port': 5432,
            'postgres_username': 'jesse_user',
            'postgres_password': 'password',
        },

        'logging': {
            'order_submission': True,
            'order_cancellation': True,
            'order_execution': True,
            'position_opened': True,
            'position_increased': True,
            'position_reduced': True,
            'position_closed': True,
            'shorter_period_candles': False,
            'trading_candles': True,
            'balance_update': True,
        },

        'exchanges': {
            'Sandbox': {
                'fee': 0,
                'starting_balance': 10000,
            },

            # https://www.bitfinex.com
            'Bitfinex': {
                'fee': 0.002,
                'starting_balance': 10000,
            },

            # https://www.binance.com
            'Binance': {
                'fee': 0.001,
                'starting_balance': 10000,
            },

            # https://www.binance.com
            'Binance Futures': {
                'fee': 0.0002,
                'starting_balance': 10000,
            },

            # https://testnet.binancefuture.com
            'Testnet Binance Futures': {
                'fee': 0.0002,
                'starting_balance': 10000,
            },

            # https://pro.coinbase.com
            'Coinbase': {
                'fee': 0.005,
                'starting_balance': 10000,
            },
        },

        # changes the metrics output of the backtest
        'metrics': {
            'sharpe_ratio': True,
            'calmar_ratio': False,
            'sortino_ratio': False,
            'omega_ratio': False,
            'winning_streak': False,
            'losing_streak': False,
            'largest_losing_trade': False,
            'largest_winning_trade': False,
            'total_winning_trades': False,
            'total_losing_trades': False,
        }
    },

    # These values are just placeholders used by Jesse at runtime
    'app': {
        # list of currencies to consider
        'considering_symbols': [],
        # The symbol to trade.
        'trading_symbols': [],

        # list of time frames to consider
        'considering_timeframes': [],
        # Which candle type do you intend trade on
        'trading_timeframes': [],

        # list of exchanges to consider
        'considering_exchanges': [],
        # list of exchanges to consider
        'trading_exchanges': [],

        'considering_candles': [],

        # dict of registered live trade drivers
        'live_drivers': {},

        # Accepted values are: 'backtest', 'livetrade', 'fitness'.
        'trading_mode': '',

        # variable used for test-driving the livetrade mode
        'is_test_driving': False,

        # this would enable many console.log()s in the code, which are helpful for debugging.
        'debug_mode': False,

        # setting this value to False would disable few validation checks which
        # are required for live trading on real markets, however, it is useful
        # for doing backtests simulations while faster strategy development.
        'strategy-validation': True
    },
}

backup_config = config.copy()


def set_config(c):
    """

    :param c:
    """
    global config
    config['env'] = c
    # add sandbox because it isn't in the local config file
    config['env']['exchanges']['Sandbox'] = {
        'fee': 0,
        'starting_balance': 10000,
    }


def reset_config():
    """

    """
    global config
    config = backup_config.copy()



from jesse.config import config
from jesse.enums import exchanges, timeframes
from jesse.routes import router
from jesse.store import store


def test_routes():
    # re-define routes
    router.set_routes([
        (exchanges.BITFINEX, 'ETHUSD', timeframes.HOUR_3, 'Test19'),
        (exchanges.SANDBOX, 'BTCUSD', timeframes.MINUTE_15, 'Test19'),
    ])

    router.set_extra_candles([
        (exchanges.BITFINEX, 'EOSUSD', timeframes.HOUR_3),
        (exchanges.BITFINEX, 'EOSUSD', timeframes.HOUR_1),
    ])

    # reset store for new routes to take affect
    store.reset(True)

    # now assert it's working as expected
    assert set(config['app']['trading_exchanges']) == set([exchanges.SANDBOX, exchanges.BITFINEX])
    assert set(config['app']['trading_symbols']) == set(['BTCUSD', 'ETHUSD'])
    assert set(config['app']['trading_timeframes']) == set([timeframes.HOUR_3, timeframes.MINUTE_15])
    assert set(config['app']['considering_exchanges']) == set([exchanges.SANDBOX, exchanges.BITFINEX])
    assert set(config['app']['considering_symbols']) == set(['BTCUSD', 'ETHUSD', 'EOSUSD'])
    assert set(config['app']['considering_timeframes']) == set(
        [timeframes.MINUTE_1, timeframes.HOUR_3, timeframes.MINUTE_15, timeframes.HOUR_1])



import peewee

import jesse.helpers as jh
from jesse.services.db import db


class Ticker(peewee.Model):
    """

    """
    id = peewee.UUIDField(primary_key=True)
    # timestamp in milliseconds
    timestamp = peewee.BigIntegerField()
    # the latest trades price
    last_price = peewee.FloatField()
    # the trading volume in the last 24 hours
    volume = peewee.FloatField()
    # the highest price in the last 24 hours
    high_price = peewee.FloatField()
    # the lowest price in the last 24 hours
    low_price = peewee.FloatField()
    symbol = peewee.CharField()
    exchange = peewee.CharField()

    class Meta:
        database = db
        indexes = ((('timestamp', 'exchange', 'symbol'), True),)

    def __init__(self, attributes=None, **kwargs):
        peewee.Model.__init__(self, attributes=attributes, **kwargs)

        if attributes is None:
            attributes = {}

        for a in attributes:
            setattr(self, a, attributes[a])


if not jh.is_unit_testing():
    # create the table
    Ticker.create_table()



def get_current_price(exchange, symbol):
    """

    :param exchange:
    :param symbol:
    :return:
    """
    p = get_position(exchange, symbol)
    return p.current_price


def get_position(exchange, symbol):
    """

    :param exchange:
    :param symbol:
    :return:
    """
    from jesse.store import store
    key = '{}-{}'.format(exchange, symbol)
    return store.positions.storage.get(key, None)


def get_orders(exchange, symbol):
    """

    :param exchange:
    :param symbol:
    :return:
    """
    from jesse.store import store
    return store.orders.get_orders(exchange, symbol)


def get_time():
    """

    :return:
    """
    from jesse.store import store
    return store.app.time


def get_exchange(name):
    """

    :param name:
    :return:
    """
    from jesse.store import store
    return store.exchanges.storage.get(name, None)


def get_strategy(exchange, symbol):
    """

    :param exchange:
    :param symbol:
    :return:
    """
    from jesse.routes import router
    r = next(r for r in router.routes
             if r.exchange == exchange and r.symbol == symbol)
    return r.strategy


def get_route(exchange, symbol):
    """

    :param exchange:
    :param symbol:
    :return:
    """
    from jesse.routes import router
    r = next(
        (r for r in router.routes if r.exchange == exchange and r.symbol == symbol),
        None)
    return r



import requests

import jesse.helpers as jh
from jesse import exceptions
from .interface import CandleExchange


class Bitfinex(CandleExchange):
    """

    """
    def __init__(self):
        super().__init__('Bitfinex', 1440, 1)
        self.endpoint = 'https://api-pub.bitfinex.com/v2/candles'

    def init_backup_exchange(self):
        from .coinbase import Coinbase
        self.backup_exchange = Coinbase()

    def get_starting_time(self, symbol: str):
        """

        :param symbol:
        :return:
        """
        # hard-code few common symbols
        if symbol == 'BTCUSD':
            return jh.date_to_timestamp('2015-08-01')
        elif symbol == 'ETHUSD':
            return jh.date_to_timestamp('2016-01-01')

        payload = {
            'sort': 1,
            'limit': 5000,
        }

        response = requests.get(self.endpoint + '/trade:1D:t{}/hist'.format(symbol), params=payload)

        if response.status_code != 200:
            raise Exception(response.content)

        data = response.json()

        # wrong symbol entered
        if not len(data):
            raise exceptions.SymbolNotFound(
                "No candle exists for {} in Bitfinex. You're probably misspelling the symbol name.".format(symbol)
            )

        first_timestamp = int(data[0][0])
        second_timestamp = first_timestamp + 60_000 * 1440

        return second_timestamp

    def fetch(self, symbol, start_timestamp):
        """

        :param symbol:
        :param start_timestamp:
        :return:
        """
        # since Bitfinex API skips candles with "volume=0", we have to send end_timestamp
        # instead of limit. Therefore, we use limit number to calculate the end_timestamp
        end_timestamp = start_timestamp + (self.count - 1) * 60000

        payload = {
            'start': start_timestamp,
            'end': end_timestamp,
            'limit': self.count,
            'sort': 1
        }

        response = requests.get(
            self.endpoint + '/trade:1m:t{}/hist'.format(symbol),
            params=payload
        )

        data = response.json()
        candles = []

        for d in data:
            candles.append({
                'id': jh.generate_unique_id(),
                'symbol': symbol,
                'exchange': self.name,
                'timestamp': d[0],
                'open': d[1],
                'close': d[2],
                'high': d[3],
                'low': d[4],
                'volume': d[5]
            })

        return candles



import os
from datetime import datetime, timedelta

import arrow
import numpy as np
import pandas as pd
from matplotlib import pyplot as plt
from pandas.plotting import register_matplotlib_converters

import jesse.helpers as jh
from jesse.config import config
from jesse.routes import router
from jesse.store import store


def portfolio_vs_asset_returns():
    """

    """
    register_matplotlib_converters()
    trades = store.completed_trades.trades
    # create a plot figure
    plt.figure(figsize=(26, 16))

    # daily balance
    plt.subplot(2, 1, 1)
    start_date = datetime.fromtimestamp(store.app.starting_time / 1000)
    date_list = [start_date + timedelta(days=x) for x in range(len(store.app.daily_balance))]
    plt.xlabel('date')
    plt.ylabel('balance')
    plt.title('Portfolio Daily Return')
    plt.plot(date_list, store.app.daily_balance)

    # price change%
    plt.subplot(2, 1, 2)
    price_dict = {}
    for r in router.routes:
        key = jh.key(r.exchange, r.symbol)
        price_dict[key] = {
            'indexes': {},
            'prices': []
        }
        dates = []
        prices = []
        candles = store.candles.get_candles(r.exchange, r.symbol, '1m')
        max_timeframe = jh.max_timeframe(config['app']['considering_timeframes'])
        pre_candles_count = jh.timeframe_to_one_minutes(max_timeframe) * 210
        for i, c in enumerate(candles):
            # do not plot prices for required_initial_candles period
            if i < pre_candles_count:
                continue

            dates.append(datetime.fromtimestamp(c[0] / 1000))
            prices.append(c[2])
            # save index of the price instead of the actual price
            price_dict[key]['indexes'][str(int(c[0]))] = len(prices) - 1

        # price => %returns
        price_returns = pd.Series(prices).pct_change(1) * 100
        cumsum_returns = np.cumsum(price_returns)
        if len(router.routes) == 1:
            plt.plot(dates, cumsum_returns, label=r.symbol, c='grey')
        else:
            plt.plot(dates, cumsum_returns, label=r.symbol)
        price_dict[key]['prices'] = cumsum_returns

    # buy and sell plots
    buy_x = []
    buy_y = []
    sell_x = []
    sell_y = []
    for t in trades:
        key = jh.key(t.exchange, t.symbol)

        if t.type == 'long':
            buy_x.append(datetime.fromtimestamp(t.opened_at / 1000))
            sell_x.append(datetime.fromtimestamp(t.closed_at / 1000))
            # add price change%
            buy_y.append(
                price_dict[key]['prices'][price_dict[key]['indexes'][str(int(t.opened_at))]]
            )
            sell_y.append(
                price_dict[key]['prices'][price_dict[key]['indexes'][str(int(t.closed_at))]]
            )
        elif t.type == 'short':
            buy_x.append(datetime.fromtimestamp(t.closed_at / 1000))
            sell_x.append(datetime.fromtimestamp(t.opened_at / 1000))
            # add price change%
            buy_y.append(
                price_dict[key]['prices'][price_dict[key]['indexes'][str(int(t.closed_at))]]
            )
            sell_y.append(
                price_dict[key]['prices'][price_dict[key]['indexes'][str(int(t.opened_at))]]
            )

    plt.plot(buy_x, buy_y, '.', color='green')
    plt.plot(sell_x, sell_y, '.', color='red')

    plt.xlabel('date')
    plt.ylabel('price change %')
    plt.title('Asset Daily Return')
    plt.legend(loc='upper left')

    # store final result
    mode = config['app']['trading_mode']
    if mode == 'backtest':
        mode = 'BT'
    if mode == 'livetrade':
        mode = 'LT'
    if mode == 'papertrade':
        mode = 'PT'

    # make sure directories exist
    os.makedirs('./storage/charts', exist_ok=True)
    file_path = 'storage/charts/{}-{}.png'.format(
        mode, str(arrow.utcnow())[0:19]
    ).replace(":", "-")
    plt.savefig(file_path)

    print('\nChart output saved at:\n{}'.format(file_path))



import numpy as np


def get_candles(exchange: str, symbol: str, timeframe: str, start_date: str, finish_date: str) -> np.ndarray:
    """
    Returns candles from the database in numpy format

    :param exchange: str
    :param symbol: str
    :param timeframe: str
    :param start_date: str
    :param finish_date: str
    
    :return: np.ndarray
    """
    exchange = exchange.title()
    symbol = symbol.upper()

    import arrow

    import jesse.helpers as jh
    from jesse.models import Candle
    from jesse.exceptions import Breaker
    from jesse.services.candle import generate_candle_from_one_minutes

    start_date = jh.arrow_to_timestamp(arrow.get(start_date, 'YYYY-MM-DD'))
    finish_date = jh.arrow_to_timestamp(arrow.get(finish_date, 'YYYY-MM-DD')) - 60000

    # validate
    if start_date == finish_date:
        raise ValueError('start_date and finish_date cannot be the same.')
    if start_date > finish_date:
        raise ValueError('start_date cannot be bigger than finish_date.')
    if finish_date > arrow.utcnow().timestamp * 1000:
        raise ValueError('Can\'t backtest the future!')

    # fetch from database
    candles_tuple = Candle.select(
        Candle.timestamp, Candle.open, Candle.close, Candle.high, Candle.low,
        Candle.volume
    ).where(
        Candle.timestamp.between(start_date, finish_date),
        Candle.exchange == exchange,
        Candle.symbol == symbol).order_by(Candle.timestamp.asc()).tuples()

    candles = np.array(tuple(candles_tuple))

    # validate that there are enough candles for selected period
    if len(candles) == 0 or candles[-1][0] != finish_date or candles[0][0] != start_date:
        raise Breaker('Not enough candles for {}. Try running "jesse import-candles"'.format(symbol))

    if timeframe == '1m':
        return candles

    generated_candles = []
    for i in range(len(candles)):
        num = jh.timeframe_to_one_minutes(timeframe)

        if (i + 1) % num == 0:
            generated_candles.append(
                generate_candle_from_one_minutes(
                    timeframe,
                    candles[(i - (num - 1)):(i + 1)],
                    True
                )
            )

    return np.array(generated_candles)



from typing import Union

import numpy as np

from jesse.helpers import get_candle_source, np_shift


def vpt(candles: np.ndarray, source_type="close", sequential=False) -> Union[float, np.ndarray]:
    """
    Volume Price Trend (VPT)

    :param candles: np.ndarray
    :param source_type: str - default: "close"
    :param sequential: bool - default=False

    :return: float | np.ndarray
    """
    if not sequential and len(candles) > 240:
        candles = candles[-240:]

    source = get_candle_source(candles, source_type=source_type)
    vpt = (candles[:, 5] * ((source - np_shift(source, 1, fill_value=np.nan)) / np_shift(source, 1, fill_value=np.nan)))
    res = np_shift(vpt, 1, fill_value=np.nan) + vpt
    
    return res if sequential else res[-1]

